--[[

Nintendo Tetris AI
Copyright (C) 2014 meatfighter.com

This file is part of Nintendo Tetris AI.

Nintendo Tetris AI is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published
by the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

Nintendo Tetris AI is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

]]

-- This Lua script works with FCEUX 2.2.2 and Nintendo Tetris (USA version).

PLAYFIELD_WIDTH = 10
PLAYFIELD_HEIGHT = 20
TETRIMINOS_SEARCHED = 2

Tetriminos = {

  NONE = -1,
  T = 1,
  J = 2,
  Z = 3,
  O = 4,
  S = 5,
  L = 6,
  I = 7,

  ORIENTATIONS = { }, 
  TYPES = { } 
}

globalMark = 1

function newQueue()

  local self = { 
    enqueue = nil,
    dequeue = nil,
    isEmpty = nil,
    isNotEmpty = nil,
  }

  local head = nil
  local tail = nil

  function self.enqueue(state)
    if head == nil then
      head = state
      tail = state        
    else
      tail.nextState = state
      tail = state        
    end
    state.nextState = nil
  end

  function self.dequeue()
    local state = head
    if head ~= nil then
      if head == tail then
        head = nil
        tail = nil
      else
        head = head.nextState
      end
    end
    return state
  end

  function self.isEmpty() 
    return head == nil
  end

  function self.isNotEmpty() 
    return head ~= nil
  end

  return self
end

function newSearcher(searchListener)

  local self = {
    search = nil
  }

  local queue = newQueue()
  local states = { }
  local lockedStates = { }
  local speed = 1

  local addChild = nil
  local createStates = nil
  local createLockedStates = nil
  local testPosition = nil
  local lockTetrimino = nil

  function testPosition(playfield, tetriminoType, x, y, rotation) 

    local orientation = Tetriminos.ORIENTATIONS[tetriminoType][rotation]
    if x < orientation.minX or x > orientation.maxX or y > orientation.maxY then
      return false
    end

    local squares = orientation.squares
    for i = 1, 4 do
      local square = squares[i]
      local rowY = y + square.y
      if rowY >= 1 
          and playfield[rowY][x + square.x] ~= Tetriminos.NONE then
        return false
      end
    end

    return true
  end

  function lockTetrimino(playfield, tetriminoType, id, state)
    local squares 
        = Tetriminos.ORIENTATIONS[tetriminoType][state.rotation].squares
    for i = 1, 4 do
      local square = squares[i]
      local y = state.y + square.y
      if y >= 1 then
        playfield[y][state.x + square.x] = tetriminoType
        playfield[y][PLAYFIELD_WIDTH + 1] 
            = playfield[y][PLAYFIELD_WIDTH + 1] + 1
      end
    end
    searchListener(playfield, tetriminoType, id, state)
    for i = 1, 4 do
      local square = squares[i]
      local y = state.y + square.y
      if y >= 1 then
        playfield[y][state.x + square.x] = Tetriminos.NONE
        playfield[y][PLAYFIELD_WIDTH + 1] 
            = playfield[y][PLAYFIELD_WIDTH + 1] - 1
      end
    end
  end

  function addChild(playfield, tetriminoType, id, maxRotation, mark, 
      level, state, l, r, d, a, b)

    local x = state.x
    local y = state.y
    local rotation = state.rotation
    local fallTimer = state.fallTimer + 1
    local rotationA = rotation == maxRotation and 1 or (rotation + 1)
    local rotationB = rotation == 1 and maxRotation or (rotation - 1)

    if state.seed then
      if fallTimer >= speed then
        if testPosition(playfield, tetriminoType, x, y + 1, rotation) then
          y = y + 1
          fallTimer = 0
        else
          local lockedState = lockedStates[y][x][rotation]
          if lockedState.visited ~= mark then
            lockedState.visited = mark
            lockedState.predecessor = state
            lockedState.l = l
            lockedState.r = r
            lockedState.d = d
            lockedState.a = a
            lockedState.b = b
            lockTetrimino(playfield, tetriminoType, id, lockedState)
          end
          return
        end
      end
      fallTimer = fallTimer + 1
    end

    if l == 2 
        and testPosition(playfield, tetriminoType, x - 1, y, rotation) then
      x = x - 1
    elseif r == 2 
        and testPosition(playfield, tetriminoType, x + 1, y, rotation) then
      x = x + 1 
    end

    if a == 2 
        and testPosition(playfield, tetriminoType, x, y, rotationA) then
      rotation = rotationA
    elseif b == 2 
        and testPosition(playfield, tetriminoType, x, y, rotationB) then
      rotation = rotationB    
    end
        
    if d == 4 or fallTimer >= speed 
        or (fallTimer < 0 and level > 9 and d > 1) then
      if testPosition(playfield, tetriminoType, x, y + 1, rotation) then
        y = y + 1
        fallTimer = 0
      else
        local lockedState = lockedStates[y][x][rotation]
        if lockedState.visited ~= mark then
          lockedState.visited = mark
          lockedState.predecessor = state
          lockedState.l = l
          lockedState.r = r
          lockedState.d = d
          lockedState.a = a
          lockedState.b = b
          lockTetrimino(playfield, tetriminoType, id, lockedState)
        end
        return
      end
    end       

    local childState = states[y][x][rotation][l][r][d][a][b]
    if childState.visited ~= mark and not childState.seed then
      childState.visited = mark
      childState.predecessor = state  
      childState.fallTimer = fallTimer  
      queue.enqueue(childState)
    end

  end

  function createStates()
    for y = 1, PLAYFIELD_HEIGHT do
      states[y] = { }
      for x = 1, PLAYFIELD_WIDTH do
        states[y][x] = { }
        for rotation = 1, 4 do 
          states[y][x][rotation] = { }
          for l = 1, 2 do
            states[y][x][rotation][l] = { }
            for r = 1, 2 do
              states[y][x][rotation][l][r] = { }
              for d = 1, 4 do
                states[y][x][rotation][l][r][d] = { }
                for a = 1, 2 do
                  states[y][x][rotation][l][r][d][a] = { }
                  for b = 1, 2 do
                    states[y][x][rotation][l][r][d][a][b] = {
                      y = y, 
                      x = x, 
                      rotation = rotation,
                      l = l, 
                      r = r, 
                      d = d, 
                      a = a, 
                      b = b,
                      visited = 0,
                      predecessor = nil,
                      fallTimer = 0,
                      nextState = nil,
                      seed = x == 6 and y == 1 and rotation == 1,
                    }
                  end
                end
              end
            end
          end
        end
      end
    end
  end

  function createLockedStates()
    for y = 1, PLAYFIELD_HEIGHT do
      lockedStates[y] = { }
      for x = 1, PLAYFIELD_WIDTH do
        lockedStates[y][x] = { }
        for rotation = 1, 4 do 
          lockedStates[y][x][rotation] = {
              y = y, 
              x = x, 
              rotation = rotation,
              l = 0, 
              r = 0, 
              d = 0, 
              a = 0, 
              b = 0,
              visited = 0,
              predecessor = nil,
              fallTimer = 0,
              nextState = nil,
              seed = false,
            }
        end
      end
    end
  end

  function self.search(playfield, tetriminoType, _speed, level, fallTimer, id)

    speed = _speed

    local maxRotation = #Tetriminos.ORIENTATIONS[tetriminoType]

    local mark = globalMark
    globalMark = globalMark + 1 

    if not testPosition(playfield, tetriminoType, 6, 1, 1) then
      return false
    end   

    local seedState = states[1][6][1][1][1][1][1][1];    
    seedState.fallTimer = fallTimer

    addChild(playfield, tetriminoType, id, maxRotation, mark, 
        level, seedState, 2, 1, 1, 1, 1)
    addChild(playfield, tetriminoType, id, maxRotation, mark,
        level, seedState, 1, 2, 1, 1, 1)
    addChild(playfield, tetriminoType, id, maxRotation, mark,
        level, seedState, 1, 1, 2, 1, 1)
    addChild(playfield, tetriminoType, id, maxRotation, mark,
        level, seedState, 2, 1, 1, 2, 1)
    addChild(playfield, tetriminoType, id, maxRotation, mark,
        level, seedState, 2, 1, 1, 1, 2)
    addChild(playfield, tetriminoType, id, maxRotation, mark,
        level, seedState, 1, 2, 1, 2, 1)
    addChild(playfield, tetriminoType, id, maxRotation, mark,
        level, seedState, 1, 2, 1, 1, 2)
    addChild(playfield, tetriminoType, id, maxRotation, mark,
        level, seedState, 1, 1, 2, 2, 1)
    addChild(playfield, tetriminoType, id, maxRotation, mark,
        level, seedState, 1, 1, 2, 1, 2)
    addChild(playfield, tetriminoType, id, maxRotation, mark,
        level, seedState, 1, 1, 1, 2, 1)
    addChild(playfield, tetriminoType, id, maxRotation, mark,
        level, seedState, 1, 1, 1, 1, 2)

    while queue.isNotEmpty() do
      local state = queue.dequeue()  
      local slide = state.l == 2 or state.r == 2
      local rotate = state.a == 2 or state.b == 2
      
      if state.d == 1 or state.d == 4 then
        local nextD = state.d == 1 and 2 or 3
        addChild(playfield, tetriminoType, id, maxRotation, mark,
            level, state, 1, 1, nextD, 1, 1)
        if not rotate then
          addChild(playfield, tetriminoType, id, maxRotation, mark,
              level, state, 1, 1, nextD, 2, 1)
          addChild(playfield, tetriminoType, id, maxRotation, mark,
              level, state, 1, 1, nextD, 1, 2)
        end       
        if slide then
          addChild(playfield, tetriminoType, id, maxRotation, 
              mark, level, state, 1, 1, 1, 1, 1)
          if not rotate then
            addChild(playfield, tetriminoType, id, maxRotation, 
                mark, level, state, 1, 1, 1, 2, 1)
            addChild(playfield, tetriminoType, id, maxRotation, 
                mark, level, state, 1, 1, 1, 1, 2)
          end        
        else
          addChild(playfield, tetriminoType, id, maxRotation, 
              mark, level, state, 2, 1, 1, 1, 1)
          addChild(playfield, tetriminoType, id, maxRotation, 
              mark, level, state, 1, 2, 1, 1, 1)
          if not rotate then
            addChild(playfield, tetriminoType, id, maxRotation, 
                mark, level, state, 2, 1, 1, 2, 1)
            addChild(playfield, tetriminoType, id, maxRotation, 
                mark, level, state, 2, 1, 1, 1, 2)
            addChild(playfield, tetriminoType, id, maxRotation, 
                mark, level, state, 1, 2, 1, 2, 1)
            addChild(playfield, tetriminoType, id, maxRotation, 
                mark, level, state, 1, 2, 1, 1, 2)
          end
        end        
      else 
        local nextD = state.d + 1        
        addChild(playfield, tetriminoType, id, maxRotation, mark,
            level, state, 1, 1, nextD, 1, 1)
        if not rotate then
          addChild(playfield, tetriminoType, id, maxRotation, mark,
              level, state, 1, 1, nextD, 2, 1)
          addChild(playfield, tetriminoType, id, maxRotation, mark,
              level, state, 1, 1, nextD, 1, 2)
        end 
      end    
    end

    return true
  end

  createStates()
  createLockedStates()

  return self
end

function newPlayfieldUtil()

  local self = {
    clearRows = nil,
    restoreRows = nil,
    evaluatePlayfield = nil,
  }

  local spareRows = { }
  local spareIndex = 1
  local columnDepths = { }

  local clearRow = nil
  local restoreRow = nil

  function clearRow(playfield, y)
    local clearedRow = playfield[y]
    clearedRow[PLAYFIELD_WIDTH + 1] = y;
    for i = y, 2, -1 do
      playfield[i] = playfield[i - 1]
    end
    playfield[1] = spareRows[spareIndex]
    playfield[1][PLAYFIELD_WIDTH + 1] = 0

    spareRows[spareIndex] = clearedRow
    spareIndex = spareIndex + 1
  end

  function restoreRow(playfield)
    spareIndex = spareIndex - 1    
    local restoredRow = spareRows[spareIndex]
    local y = restoredRow[PLAYFIELD_WIDTH + 1]

    spareRows[spareIndex] = playfield[1]
    for i = 1, y - 1 do
      playfield[i] = playfield[i + 1]
    end
    restoredRow[PLAYFIELD_WIDTH + 1] = PLAYFIELD_WIDTH
    playfield[y] = restoredRow    
  end

  function self.evaluatePlayfield(playfield, e)

    for x = 1, PLAYFIELD_WIDTH do
      for y = 1, PLAYFIELD_HEIGHT do
        if y == PLAYFIELD_HEIGHT or playfield[y][x] ~= Tetriminos.NONE then
          columnDepths[x] = y
          break
        end
      end
    end

    e.wells = 0
    for x = 1, PLAYFIELD_WIDTH do
      local minY = 1
      if x == 1 then
        minY = columnDepths[2]
      elseif x == PLAYFIELD_WIDTH then
        minY = columnDepths[PLAYFIELD_WIDTH - 1]
      else
        minY = math.max(columnDepths[x - 1],
            columnDepths[x + 1])
      end
      for y = columnDepths[x], minY, -1 do
        if (x == 1 or playfield[y][x - 1] ~= Tetriminos.NONE)
            and (x == PLAYFIELD_WIDTH
                or playfield[y][x + 1] ~= Tetriminos.NONE) then
          e.wells = e.wells + 1
        end
      end
    end

    e.holes = 0
    e.columnTransitions = 0
    for x = 1, PLAYFIELD_WIDTH do
      local solid = true
      for y = columnDepths[x] + 1, PLAYFIELD_HEIGHT do
        if playfield[y][x] == Tetriminos.NONE then
          if playfield[y - 1][x] ~= Tetriminos.NONE then
            e.holes = e.holes + 1
          end
          if solid then
            solid = false
            e.columnTransitions = e.columnTransitions + 1
          end
        elseif not solid then
          solid = true
          e.columnTransitions = e.columnTransitions + 1
        end
      end
    end

    e.rowTransitions = 0
    for y = 1, PLAYFIELD_HEIGHT do
      local solidFound = false
      local solid = true
      local transitions = 0
      for x = 1, PLAYFIELD_WIDTH + 1 do
        if x == PLAYFIELD_WIDTH + 1 then
          if not solid then
            transitions = transitions + 1
          end
        elseif playfield[y][x] == Tetriminos.NONE then
          if solid then
            solid = false
            transitions = transitions + 1
          end
        else
          solidFound = true
          if not solid then
            solid = true
            transitions = transitions + 1
          end
        end
      end
      if solidFound then
        e.rowTransitions = e.rowTransitions + transitions
      end
    end
  end

  function self.restoreRows(playfield, rows)
    for i = 1, rows do
      restoreRow(playfield)
    end
  end

  function self.clearRows(playfield, tetriminoY)

    local rows = 0
    local startRow = tetriminoY - 2
    local endRow = tetriminoY + 1

    if startRow < 2 then
      startRow = 2
    end
    if endRow > 20 then
      endRow = 20
    end

    for y = startRow, endRow do
      if playfield[y][PLAYFIELD_WIDTH + 1] == PLAYFIELD_WIDTH then
        rows = rows + 1
        clearRow(playfield, y)
      end
    end

    return rows
  end

  for y = 1, 8 * TETRIMINOS_SEARCHED do
    spareRows[y] = { }
    for x = 1, PLAYFIELD_WIDTH do
      spareRows[y][x] = Tetriminos.NONE
    end
    spareRows[y][PLAYFIELD_WIDTH + 1] = 0
  end

  for i = 1, PLAYFIELD_WIDTH do
    columnDepths[i] = 0
  end

  return self
end

function newPlayfieldEvaluation()
  return {
    holes = 0,
    columnTransitions = 0,
    rowTransitions = 0,
    wells = 0,
  }
end

function newAI(tetriminos)

  local WEIGHTS = {
    1.0,
    12.885008263218383,
    15.842707182438396,
    26.89449650779595,
    27.616914062397015,
    30.18511071927904,
  }

  local self = {
    search = nil
  }

  local searchers = { }
  local tetriminoIndices = { }
  local playfieldUtil = newPlayfieldUtil()
  local e = newPlayfieldEvaluation()
  local totalRows = 0
  local totalDropHeight = 0
  local bestResult = nil
  local bestFitness = 0
  local result1 = nil
  local speed = 0
  local level = 0

  local computeFitness = nil
  local copyMoves = nil
  local handleResult = nil

  function computeFitness()
    return WEIGHTS[1] * totalRows
         + WEIGHTS[2] * totalDropHeight
         + WEIGHTS[3] * e.wells
         + WEIGHTS[4] * e.holes                 
         + WEIGHTS[5] * e.columnTransitions
         + WEIGHTS[6] * e.rowTransitions
  end

  function handleResult(playfield, tetriminoType, id, state)

    if id == 1 then
      result1 = state
    end

    local orientation = Tetriminos.ORIENTATIONS[tetriminoType][state.rotation]
    local rows = playfieldUtil.clearRows(playfield, state.y)
    local originalTotalRows = totalRows
    local originalTotalDropHeight = totalDropHeight
    totalRows = totalRows + rows
    totalDropHeight = totalDropHeight + orientation.maxY - state.y

    local nextID = id + 1

    if nextID == #tetriminoIndices + 1
        or not searchers[nextID].search(
            playfield, tetriminoIndices[nextID], 
                speed, level, 0, nextID) then

      playfieldUtil.evaluatePlayfield(playfield, e)

      local fitness = computeFitness()
      if fitness < bestFitness then
        bestFitness = fitness
        bestResult = result1
      end
    end

    totalDropHeight = originalTotalDropHeight
    totalRows = originalTotalRows
    playfieldUtil.restoreRows(playfield, rows)
  end

  function self.search(playfield, _tetriminoIndices, _speed, _level, fallTimer)

    tetriminoIndices = _tetriminoIndices
    speed = _speed
    level = _level
    bestResult = nil
    bestFitness = 1000000000

    searchers[1].search(playfield, tetriminoIndices[1], 
        speed, level, fallTimer, 1)

    return bestResult
  end

  for i = 1, tetriminos do
    searchers[i] = newSearcher(handleResult)
  end

  return self
end

function createEmptyPlayfield()
  local playfield = { }
  for y = 1, PLAYFIELD_HEIGHT do
    playfield[y] = { }
    for x = 1, PLAYFIELD_WIDTH do
      playfield[y][x] = Tetriminos.NONE
    end
    playfield[y][PLAYFIELD_WIDTH + 1] = 0
  end
  return playfield
end

-- FCEUX code starts

orientationTableAddress = 0x8A9C
tetriminoTypeTableAddress = 0x993B
spawnTableAddress = 0x9956
copyrightAddress1 = 0x00C3
copyrightAddress2 = 0x00A8
gameStateAddress = 0x00C0
playStateAddress = 0x0048
lowCounterAddress = 0x00B1
highCounterAddress = 0x00B2
tetriminoXAddress = 0x0060
tetriminoYAddress1 = 0x0061
tetriminoYAddress2 = 0x0041
tetriminoIDAddress = 0x0062
nextTetriminoIDAddress = 0x00BF
fallTimerAddress = 0x0065
playfieldAddress = 0x0400
levelAddress = 0x0064
levelTableAccessAddress = 0x9808
linesHighAddress = 0x0071
linesLowAddress = 0x0070

emptySquare = 0xEF

lastPlayState = 8
targetTetriminoY = 0
maxScoreHit = false
startCounter = 0

function buildOrientationTable() 
  for i = 1, 7 do
    Tetriminos.ORIENTATIONS[i] = { }
  end

  local spawnIndices = { }

  for orientationID = 0, 18 do

    local tetriminoType 
        = memory.readbyteunsigned(tetriminoTypeTableAddress + orientationID) + 1
    table.insert(Tetriminos.TYPES, tetriminoType)

    local orientation = {
      squares = { },
      minX = 100,
      maxX = -100,
      maxY = -100,
    }
    table.insert(Tetriminos.ORIENTATIONS[tetriminoType], orientation)

    if orientationID 
        == memory.readbyteunsigned(spawnTableAddress + orientationID) then
      table.insert(spawnIndices, #Tetriminos.ORIENTATIONS[tetriminoType])
    end

    for square = 0, 3 do
      local index = 12 * orientationID + 3 * square
      local x = memory.readbytesigned(orientationTableAddress + index + 2)
      local y = memory.readbytesigned(orientationTableAddress + index)
      table.insert(orientation.squares, { x = x, y = y })
      orientation.minX = math.min(orientation.minX, x)
      orientation.maxX = math.max(orientation.maxX, x)
      orientation.maxY = math.max(orientation.maxY, y)
    end

    orientation.minX = 1 - orientation.minX
    orientation.maxX = PLAYFIELD_WIDTH - orientation.maxX
    orientation.maxY = PLAYFIELD_HEIGHT - orientation.maxY
  end

  for i = 1, 7 do
    local orientations = Tetriminos.ORIENTATIONS[i]
    Tetriminos.ORIENTATIONS[i] = { }
    for j = 1, #orientations do
      table.insert(Tetriminos.ORIENTATIONS[i], 
          orientations[((j + spawnIndices[i] - 2) % #orientations) + 1])
    end
  end
end

function pressStart()
  if startCounter == 0 then
    startCounter = 10
  end
  if startCounter >= 5 then
    joypad.set(1, { start = true })
  elseif startCounter < 5 then
    joypad.set(1, { start = false })
  end
  startCounter = startCounter - 1
end

function skipCopyrightScreen(gameState)
  local copyright1 = memory.readbyteunsigned(copyrightAddress1)
  local copyright2 = memory.readbyteunsigned(copyrightAddress2)

  if gameState == 0 then
    if copyright1 > 1 then
      memory.writebyte(copyrightAddress1, 0)
    elseif copyright2 > 2 then
      memory.writebyte(copyrightAddress2, 1)
    end
  end
end

function skipTitleAndDemoScreens(gameState)
  if gameState == 1 or gameState == 5 then
    pressStart()
  end
end

function spawned()  
  local playState = memory.readbyteunsigned(playStateAddress)
  local tetriminoX = memory.readbyteunsigned(tetriminoXAddress)
  local tetriminoY = memory.readbyteunsigned(tetriminoYAddress1)
  local result = lastPlayState ~= 1 and playState == 1 and tetriminoX == 5
      and tetriminoY == 0
  lastPlayState = playState
  return result
end

function readPlayfield(playfield)
  for i = 0, 19 do
    playfield[i + 1][11] = 0
    for j = 0, 9 do
      if memory.readbyteunsigned(playfieldAddress + 10 * i + j) 
          == emptySquare then
        playfield[i + 1][j + 1] = Tetriminos.NONE
      else
        playfield[i + 1][j + 1] = Tetriminos.I
        playfield[i + 1][11] = playfield[i + 1][11] + 1
      end
    end
  end
end

function readTetrimino()
  return Tetriminos.TYPES[memory.readbyteunsigned(tetriminoIDAddress) + 1]
end

function readNextTetrimino()
  return Tetriminos.TYPES[memory.readbyteunsigned(nextTetriminoIDAddress) + 1]
end

function resetPlayState(gameState)
  if gameState ~= 4 then
    memory.writebyte(playStateAddress, 0)
  end
end

function isPlaying(gameState)
  return gameState == 4 and memory.readbyteunsigned(playStateAddress) < 9
end

function readSpeed() 
  local level = memory.readbyteunsigned(0x0044)
  if (level >= 29) then
    return 1
  else
    return memory.readbyteunsigned(0x898E + level)
  end
end

function search(tetriminos, playfield, ai)

  tetriminos[1] = readTetrimino()
  tetriminos[2] = readNextTetrimino()
  local speed = readSpeed()
  local level = memory.readbyteunsigned(0x0064)
  local fallTimer = memory.readbyteunsigned(0x0045)

  if fallTimer == 1 then
    fallTimer = -95
  end
  readPlayfield(playfield)

  local state = ai.search(playfield, tetriminos,
      speed, level, fallTimer)

  local result = { }
  if state ~= nil then
    repeat
      table.insert(result, 1, state)
      state = state.predecessor    
    until state.seed
  end

  return result
end

function updateScore()
  if memory.readbyteunsigned(0x00A8) > 30 then
    memory.writebyte(0x00A8, 30)
  end
end

function makeMove(state)
  joypad.write(1, { 
      down = state.d > 1, 
      left = state.l == 2, 
      right = state.r == 2, 
      A = state.a == 2, 
      B = state.b == 2 
  } )
end

function releaseButtons()
  joypad.write(1, { 
    down = false, 
    left = false, 
    right = false, 
    A = false, 
    B = false, 
  } )
end

do
  local tetriminos = { 0, 0 }
  local playfield = createEmptyPlayfield()
  local ai = newAI(2)
  local movesIndex = 1
  local moves = { }

  buildOrientationTable()

  memory.registerexec(0x9C35, updateScore)

  while true do

    local gameState = memory.readbyteunsigned(gameStateAddress)
    skipCopyrightScreen(gameState)
    skipTitleAndDemoScreens(gameState)
    resetPlayState(gameState)

    if (isPlaying(gameState)) then
      if movesIndex <= #moves + 1 then
        if movesIndex == #moves + 1 then
          releaseButtons()
        else
          makeMove(moves[movesIndex]) 
        end
        movesIndex = movesIndex + 1
      else
        if spawned() then 
          lastPlayState = 8         
          moves = search(tetriminos, playfield, ai)
          movesIndex = 1
        end
      end
    end

    emu.frameadvance()
  end
end

-- FCEUX code ends
